import numpy as np

style = '''
QWidget {
    font: medium Ubuntu;
    background-color: #011F2F;
    font-size: 16px;
    font-size: 16px;
    color:#FFFFFF;
}
'''
#
Nroll = 6

#  Format of the data.
forms = {
    "mag": lambda w, ang: np.abs(w),
    "phase": lambda w, ang: np.angle(w),
    "real": lambda w, ang: np.real(w),
    "imag": lambda w, ang: np.imag(w),
    "rot": lambda w, ang: np.real(np.exp(1j * ang) * np.array(w))
}
form_keys = list(forms.keys())
#
COL_SEL = (0, 0, 0)
COL_UNSEL = (6, 6, 8)
#

defualt_colors = [
    (200, 0, 0),
    (55, 100, 180),
    (40, 80, 150),
    (30, 50, 110),
    (25, 40, 70),
    (25, 30, 50),
]

widths = [3, 2, 2, 2, 1, 1]
SymSize = [5, 0, 0, 0, 0, 0]
ridx = list(range(Nroll))
ridx.reverse()



import ast
import asyncio
import inspect
import platform
import re
import subprocess
import sys
import uuid
import warnings
from typing import Any, Callable

import dill

from .expression import Env, Expression


class Unpicklable:

    def __init__(self, obj):
        self.type = str(type(obj))
        self.id = id(obj)

    def __repr__(self):
        return f'<Unpicklable: {self.type} at 0x{id(self):x}>'


class TooLarge:

    def __init__(self, obj):
        self.type = str(type(obj))
        self.id = id(obj)

    def __repr__(self):
        return f'<TooLarge: {self.type} at 0x{id(self):x}>'


def dump_dict(d, keys=[]):
    ret = {}

    for key, value in d.items():
        if key in keys:
            ret[key] = value
            continue
        if isinstance(value, dict) and isinstance(key, str):
            ret[key] = dump_dict(value,
                                 keys=[
                                     k[len(key) + 1:] for k in keys
                                     if k.startswith(f'{key}.')
                                 ])
        else:
            try:
                ret[key] = dill.dumps(value)
            except:
                ret[key] = Unpicklable(value)

    return dill.dumps(ret)


def load_dict(buff):
    if isinstance(buff, dict):
        return {key: load_dict(value) for key, value in buff.items()}

    if not isinstance(buff, bytes):
        return buff

    try:
        ret = dill.loads(buff)
    except:
        return buff

    if isinstance(ret, dict):
        return load_dict(ret)
    else:
        return ret


def dump_globals(ns=None, *, size_limit=10 * 1024 * 1024, warn=False):
    import __main__

    if ns is None:
        ns = __main__.__dict__

    namespace = {}

    for name, value in ns.items():
        try:
            buf = dill.dumps(value)
        except:
            namespace[name] = Unpicklable(value)
            if warn:
                warnings.warn(f'Unpicklable: {name} {type(value)}')
        if len(buf) > size_limit:
            namespace[name] = TooLarge(value)
            if warn:
                warnings.warn(f'TooLarge: {name} {type(value)}')
        else:
            namespace[name] = buf

    return namespace


def is_valid_identifier(s: str) -> bool:
    """
    Check if a string is a valid identifier.
    """
    try:
        ast.parse(f"f({s}=0)")
        return True
    except SyntaxError:
        return False


async def async_next(aiter):
    try:
        if hasattr(aiter, '__anext__'):
            return await aiter.__anext__()
        else:
            return next(aiter)
    except StopIteration:
        raise StopAsyncIteration from None


async def async_zip(*aiters):
    aiters = [
        ait.__aiter__() if hasattr(ait, '__aiter__') else iter(ait)
        for ait in aiters
    ]
    try:
        while True:
            # 使用 asyncio.gather 等待所有异步生成器返回下一个元素
            result = await asyncio.gather(*(async_next(ait) for ait in aiters))
            yield tuple(result)
    except StopAsyncIteration:
        # 当任一异步生成器耗尽时停止迭代
        return


async def call_function(func: Callable | Expression, variables: dict[str,
                                                                     Any]):
    if isinstance(func, Expression):
        env = Env()
        for name in func.symbols():
            if name in variables:
                if inspect.isawaitable(variables[name]):
                    variables[name] = await variables[name]
                env.variables[name] = variables[name]
            else:
                raise ValueError(f'{name} is not provided.')
        return func.eval(env)

    try:
        sig = inspect.signature(func)
    except:
        return func()
    args = []
    for name, param in sig.parameters.items():
        if param.kind == param.POSITIONAL_OR_KEYWORD:
            if name in variables:
                if inspect.isawaitable(variables[name]):
                    variables[name] = await variables[name]
                args.append(variables[name])
            elif param.default is not param.empty:
                args.append(param.default)
            else:
                raise ValueError(f'parameter {name} is not provided.')
        elif param.kind == param.VAR_POSITIONAL:
            raise ValueError('not support VAR_POSITIONAL')
        elif param.kind == param.VAR_KEYWORD:
            ret = func(**variables)
            if inspect.isawaitable(ret):
                ret = await ret
            return ret
    ret = func(*args)
    if inspect.isawaitable(ret):
        ret = await ret
    return ret


def yapf_reformat(cell_text):
    try:
        import isort
        import yapf.yapflib.yapf_api

        fname = f"f{uuid.uuid1().hex}"

        def wrap(source):
            lines = [f"async def {fname}():"]
            for line in source.split('\n'):
                lines.append("    " + line)
            return '\n'.join(lines)

        def unwrap(source):
            lines = []
            for line in source.split('\n'):
                if line.startswith(f"async def {fname}():"):
                    continue
                lines.append(line[4:])
            return '\n'.join(lines)

        cell_text = re.sub('^%', '#%#', cell_text, flags=re.M)
        reformated_text = unwrap(
            yapf.yapflib.yapf_api.FormatCode(wrap(isort.code(cell_text)))[0])
        return re.sub('^#%#', '%', reformated_text, flags=re.M)
    except:
        return cell_text


def get_installed_packages():
    result = subprocess.run([sys.executable, '-m', 'pip', 'freeze'],
                            stdout=subprocess.PIPE,
                            text=True)

    lines = result.stdout.split('\n')
    packages = []
    for line in lines:
        if line:
            packages.append(line)
    return packages


def get_system_info():
    info = {
        'OS': platform.uname()._asdict(),
        'Python': sys.version,
        'PythonExecutable': sys.executable,
        'PythonPath': sys.path,
        'packages': get_installed_packages()
    }
    return info



import warnings
from multiprocessing import Queue

from .dataset import Dataset
from .toolbar import ToolBar


class EventQueue():

    def __init__(self, queue: Queue, toolbar: ToolBar, point_dataset: Dataset,
                 trace_dataset: Dataset):
        self.toolbar = toolbar
        self.queue = queue
        self.point_dataset = point_dataset
        self.trace_dataset = trace_dataset

    def flush(self):
        while (not self.queue.empty()):
            words = self.queue.get()
            try:
                assert (isinstance(
                    words, tuple)), "QueueHandler - fifo Content must be tuple"
                assert (
                    len(words) == 2
                ), "QueueHandler -the tuple must be like (\"command_type\" , \"DATA\")"
                cmd, data = words
                assert (isinstance(
                    cmd, str)), "QueueHandler - the command should be a string"
            except AssertionError as e:
                warnings.warn(e.args[0])
                continue

            self.handle(cmd, data)

    def handle(self, cmd, data):
        match cmd:
            case "PN":
                self.point_dataset.set_column_names(data)
                self.toolbar.refresh_comb()
            case "TN":
                self.trace_dataset.set_column_names(data)
                self.toolbar.refresh_comb()
            case "PD":
                self.point_dataset.append(data)
            case "TD":
                self.trace_dataset.append(data)
            case "ROLL":
                self.point_dataset.roll()
            case "PXY":
                self.toolbar.set_point_text(data)
            case "TXY":
                self.toolbar.set_trace_text(data)
            case _:
                warnings.warn(f"QueueHandler - unknown command : {cmd}")



import functools
import pickle
import uuid
from datetime import datetime, timedelta
from pathlib import Path

from loguru import logger

from .load import WorkflowType, get_dependents
from .storage import (Result, find_result, get_heads, renew_result,
                      revoke_result, save_result)
from .transform import current_config, obey_the_oracle, update_parameters


class CalibrationFailedError(Exception):
    pass


def is_pickleable(obj) -> bool:
    try:
        pickle.dumps(obj)
        return True
    except:
        return False


def veryfy_analyzed_result(result: Result, script: str, method: str):
    if not isinstance(result, Result):
        raise TypeError(f'"{script}" : "{method}" must return a Result object')
    if not is_pickleable(result.parameters):
        raise TypeError(
            f'"{script}" : "{method}" return not pickleable data in .parameters'
        )
    if not is_pickleable(result.other_infomation):
        raise TypeError(
            f'"{script}" : "{method}" return not pickleable data in .other_infomation'
        )


def check_state(workflow: WorkflowType, code_path: str | Path,
                state_path: str | Path) -> bool:
    """
    check state should report a pass if and only if the following are satisfied:
    
    1. The cal has had check data or calibrate pass within the timeout period.
    2. The cal has not failed calibrate without resolution.
    3. No dependencies have been recalibrated since the last time check data or calibrate was run on this cal.
    4. All dependencies pass check state.
    """
    logger.debug(f'check_state: "{workflow.__workflow_id__}"')
    result = find_result(workflow.__workflow_id__, state_path)
    if not result:
        logger.debug(
            f'check_state failed: No history found for "{workflow.__workflow_id__}"'
        )
        return False
    if hasattr(workflow, 'check_state') and callable(workflow.check_state):
        logger.debug(
            f'check_state: "{workflow.__workflow_id__}" has custom check_state method'
        )
        return workflow.check_state(result)
    if workflow.__timeout__ is not None and datetime.now(
    ) > result.checked_time + timedelta(seconds=workflow.__timeout__):
        logger.debug(
            f'check_state failed: "{workflow.__workflow_id__}" has expired')
        return False
    if not result.in_spec:
        logger.debug(
            f'check_state failed: "{workflow.__workflow_id__}" is out of spec')
        return False
    if result.bad_data:
        logger.debug(
            f'check_state failed: "{workflow.__workflow_id__}" has bad data')
        return False
    for n in get_dependents(workflow, code_path):
        r = find_result(n.__workflow_id__, state_path)
        if r is None or r.checked_time > result.checked_time:
            logger.debug(
                f'check_state failed: "{workflow.__workflow_id__}" has outdated dependencies'
            )
            return False
    for n in get_dependents(workflow, code_path):
        if not check_state(n, code_path, state_path):
            logger.debug(
                f'check_state failed: "{workflow.__workflow_id__}" has bad dependencies'
            )
            return False
    return True


def call_analyzer(node,
                  result: Result,
                  history: Result | None,
                  check=False,
                  plot=False) -> Result:
    if check:
        result = node.check_analyze(result, history=history)
        veryfy_analyzed_result(result, node.__workflow_id__, "check_analyze")
        result.fully_calibrated = False
    else:
        result = node.analyze(result, history=history)
        veryfy_analyzed_result(result, node.__workflow_id__, "analyze")
        if hasattr(node, 'oracle') and callable(node.oracle):
            logger.debug(
                f'"{node.__workflow_id__}" has oracle method, calling ...')
            try:
                result = node.oracle(result,
                                     history=history,
                                     system_state=get_heads(result.base_path))
            except Exception as e:
                logger.exception(e)
                result.oracle = {}
            if not isinstance(result, Result):
                raise TypeError(
                    f'"{node.__workflow_id__}" : function "oracle" must return a Result object'
                )
            if not is_pickleable(result.oracle):
                raise TypeError(
                    f'"{node.__workflow_id__}" : function "oracle" return not pickleable data'
                )
        result.fully_calibrated = True
        if plot:
            call_plot(node, result)
    return result


@logger.catch()
def call_plot(node, result: Result, check=False):
    if hasattr(node, 'plot') and callable(node.plot):
        node.plot(result)


@functools.lru_cache(maxsize=128)
def check_data(workflow: WorkflowType, code_path: str | Path,
               state_path: str | Path, plot: bool, session_id: str) -> Result:
    """
    check data answers two questions:
    Is the parameter associated with this cal in spec,
    and is the cal scan working as expected?
    """
    history = find_result(workflow.__workflow_id__, state_path)

    if history is None:
        logger.debug(f'No history found for "{workflow.__workflow_id__}"')
        result = Result(workflow=workflow.__workflow_id__,
                        config_path=current_config(state_path),
                        base_path=state_path)
        result.in_spec = False
        result.bad_data = False
        return result

    if history.bad_data:
        logger.debug(
            f'History found for "{workflow.__workflow_id__}", but bad data')
        return history
    if not history.in_spec:
        logger.debug(
            f'History found for "{workflow.__workflow_id__}", but out of spec')
        return history

    logger.debug(
        f'History found for "{workflow.__workflow_id__}", but has expired')

    if hasattr(workflow, 'check') and callable(workflow.check) and hasattr(
            workflow, 'check_analyze') and callable(workflow.check_analyze):
        logger.debug(
            f'Checking "{workflow.__workflow_id__}" with "check" method ...')
        data = workflow.check()
        if not is_pickleable(data):
            raise TypeError(
                f'"{workflow.__workflow_id__}" : "check" return not pickleable data'
            )
        result = Result(workflow=workflow.__workflow_id__,
                        data=data,
                        config_path=current_config(state_path),
                        base_path=state_path)
        #save_result(workflow.__workflow_id__, result, state_path)

        logger.debug(f'Checked "{workflow.__workflow_id__}" !')
        result = call_analyzer(workflow,
                               result,
                               history,
                               check=True,
                               plot=plot)
        if result.in_spec:
            logger.debug(
                f'"{workflow.__workflow_id__}": checked in spec, renewing result'
            )
            renew_result(workflow.__workflow_id__, state_path)
        else:
            logger.debug(
                f'"{workflow.__workflow_id__}": checked out of spec, revoking result'
            )
            revoke_result(workflow.__workflow_id__, state_path)
    else:
        logger.debug(
            f'Checking "{workflow.__workflow_id__}" with "calibrate" method ...'
        )
        data = workflow.calibrate()
        if not is_pickleable(data):
            raise TypeError(
                f'"{workflow.__workflow_id__}" : "calibrate" return not pickleable data'
            )
        result = Result(workflow=workflow.__workflow_id__,
                        data=data,
                        config_path=current_config(state_path),
                        base_path=state_path)
        save_result(workflow.__workflow_id__, result, state_path)

        logger.debug(f'Calibrated "{workflow.__workflow_id__}" !')
        result = call_analyzer(workflow,
                               result,
                               history,
                               check=False,
                               plot=plot)
        save_result(workflow.__workflow_id__,
                    result,
                    state_path,
                    overwrite=True)

    return result


@functools.lru_cache(maxsize=128)
def calibrate(workflow: WorkflowType, code_path: str | Path,
              state_path: str | Path, plot: bool, session_id: str) -> Result:
    history = find_result(workflow.__workflow_id__, state_path)

    logger.debug(f'Calibrating "{workflow.__workflow_id__}" ...')
    data = workflow.calibrate()
    if not is_pickleable(data):
        raise TypeError(
            f'"{workflow.__workflow_id__}" : "calibrate" return not pickleable data'
        )
    result = Result(workflow=workflow.__workflow_id__,
                    data=data,
                    config_path=current_config(state_path),
                    base_path=state_path)
    save_result(workflow.__workflow_id__, result, state_path)
    logger.debug(f'Calibrated "{workflow.__workflow_id__}" !')
    result = call_analyzer(workflow, result, history, check=False, plot=plot)
    save_result(workflow.__workflow_id__, result, state_path, overwrite=True)
    return result


def diagnose(workflow: WorkflowType, code_path: str | Path,
             state_path: str | Path, plot: bool, session_id: str):
    '''
    Returns: True if node or dependent recalibrated.
    '''
    logger.debug(f'diagnose "{workflow.__workflow_id__}"')
    # check_data
    result = check_data(workflow, code_path, state_path, plot, session_id)
    # in spec case
    if result.in_spec:
        logger.debug(
            f'"{workflow.__workflow_id__}": Checked! In spec, no need to diagnose'
        )
        return False
    # bad data case
    recalibrated = []
    if result.bad_data:
        logger.debug(
            f'"{workflow.__workflow_id__}": Bad data, diagnosing dependents')
        recalibrated = [
            diagnose(n, code_path, state_path, plot, session_id)
            for n in get_dependents(workflow, code_path)
        ]
    if not any(recalibrated):
        if result.bad_data:
            raise CalibrationFailedError(
                f'"{workflow.__workflow_id__}": bad data but no dependents recalibrated.'
            )
        logger.debug(
            f'"{workflow.__workflow_id__}": no dependents recalibrated.')
    # calibrate
    if any(recalibrated):
        logger.debug(
            f'recalibrate "{workflow.__workflow_id__}" because some dependents recalibrated.'
        )
    elif not result.in_spec and not result.bad_data:
        logger.debug(
            f'recalibrate "{workflow.__workflow_id__}" because out of spec.')
    elif result.in_spec:
        logger.error(
            f'Never reach: recalibrate "{workflow.__workflow_id__}" because in spec.'
        )
    elif result.bad_data:
        logger.error(
            f'Never reach: recalibrate "{workflow.__workflow_id__}" because bad data.'
        )
    else:
        logger.error(f'Never reach: recalibrate "{workflow.__workflow_id__}"')

    result = calibrate(workflow, code_path, state_path, plot, session_id)
    if result.bad_data or not result.in_spec:
        obey_the_oracle(result, state_path)
        raise CalibrationFailedError(
            f'"{workflow.__workflow_id__}": All dependents passed, but calibration failed!'
        )
    update_parameters(result, state_path)
    return True


@logger.catch(reraise=True)
def maintain(workflow: WorkflowType,
             code_path: str | Path,
             state_path: str | Path,
             session_id: str | None = None,
             run: bool = False,
             plot: bool = False,
             freeze: bool = False):
    if session_id is None:
        session_id = uuid.uuid4().hex
    logger.debug(f'run "{workflow.__workflow_id__}"'
                 if run else f'maintain "{workflow.__workflow_id__}"')
    # recursive maintain
    for n in get_dependents(workflow, code_path):
        logger.debug(
            f'maintain "{n.__workflow_id__}" because it is depended by "{workflow.__workflow_id__}"'
        )
        maintain(n,
                 code_path,
                 state_path,
                 session_id,
                 run=False,
                 plot=plot,
                 freeze=freeze)
    else:
        logger.debug(
            f'"{workflow.__workflow_id__}": All dependents maintained')
    # check_state
    if check_state(workflow, code_path, state_path) and not run:
        logger.debug(
            f'"{workflow.__workflow_id__}": In spec, no need to maintain')
        return
    # check_data
    result = check_data(workflow, code_path, state_path, plot, session_id)
    if result.in_spec:
        if not run:
            logger.debug(
                f'"{workflow.__workflow_id__}": In spec, no need to maintain')
            return
    elif result.bad_data:
        logger.debug(
            f'"{workflow.__workflow_id__}": Bad data, diagnosing dependents')
        for n in get_dependents(workflow, code_path):
            logger.debug(
                f'diagnose "{n.__workflow_id__}" because of "{workflow.__workflow_id__}" bad data'
            )
            diagnose(n, code_path, state_path, plot, session_id)
        else:
            logger.debug(
                f'"{workflow.__workflow_id__}": All dependents diagnosed')
    # calibrate
    logger.debug(f'recalibrate "{workflow.__workflow_id__}"')
    result = calibrate(workflow, code_path, state_path, plot, session_id)
    if result.bad_data or not result.in_spec:
        if not freeze:
            obey_the_oracle(result, state_path)
        raise CalibrationFailedError(
            f'"{workflow.__workflow_id__}": All dependents passed, but calibration failed!'
        )
    if not freeze:
        update_parameters(result, state_path)
    return


@logger.catch(reraise=True)
def run(workflow: WorkflowType,
        code_path: str | Path,
        state_path: str | Path,
        plot: bool = False,
        freeze: bool = False):
    session_id = uuid.uuid4().hex
    logger.debug(f'run "{workflow.__workflow_id__}" without dependences.')
    result = calibrate(workflow,
                       code_path,
                       state_path,
                       plot,
                       session_id=session_id)
    if result.bad_data or not result.in_spec:
        if not freeze:
            obey_the_oracle(result, state_path)
        raise CalibrationFailedError(
            f'"{workflow.__workflow_id__}": All dependents passed, but calibration failed!'
        )
    if not freeze:
        update_parameters(result, state_path)
    return



import hashlib
from datetime import datetime

from sqlalchemy import (Column, DateTime, ForeignKey, Integer, LargeBinary,
                        String, Text)
from sqlalchemy.orm import relationship

from .base import Base


class InputText(Base):
    __tablename__ = 'inputs'

    id = Column(Integer, primary_key=True)
    hash = Column(LargeBinary(20))
    text_field = Column(Text, unique=True)

    @property
    def text(self):
        return self.text_field

    @text.setter
    def text(self, text):
        self.hash = hashlib.sha1(text.encode('utf-8')).digest()
        self.text_field = text

    def __repr__(self) -> str:
        return self.text


class Cell(Base):
    __tablename__ = 'cells'

    id = Column(Integer, primary_key=True)
    notebook_id = Column(Integer, ForeignKey("notebooks.id"))
    index = Column(Integer)
    ctime = Column(DateTime, default=datetime.utcnow)
    ftime = Column(DateTime, default=datetime.utcnow)
    input_id = Column(Integer, ForeignKey("inputs.id"))

    notebook = relationship("Notebook", back_populates="cells")
    input = relationship("InputText")

    def __repr__(self) -> str:
        return f"Cell(index={self.index}, input='{self.input}')"


class Notebook(Base):
    __tablename__ = 'notebooks'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    ctime = Column(DateTime, default=datetime.utcnow)
    atime = Column(DateTime, default=datetime.utcnow)

    cells = relationship("Cell",
                         order_by=Cell.index,
                         back_populates="notebook")



import lzma
import pathlib

import dill
import numpy as np

MAGIC = b'wAvEDatA'

lzma_filters = [{
    'id': lzma.FILTER_LZMA2,
    'preset': 9 | lzma.PRESET_EXTREME,
}]

FORMAT_RAW = 1
FORMAT_XZ = 2


def _get_header(file):
    assert MAGIC == file.read(len(MAGIC)), 'Invalid file format'
    version = file.read(1)
    version = int.from_bytes(version, 'big', signed=False)
    file_format = file.read(1)
    file_format = int.from_bytes(file_format, 'big', signed=False)
    info_size = file.read(2)
    info_size = int.from_bytes(info_size, 'big', signed=False)
    header = lzma.decompress(file.read(info_size),
                             format=lzma.FORMAT_RAW,
                             filters=lzma_filters)
    info = dill.loads(header)
    return {'version': version, 'format': file_format, 'info': info}


def _make_header(header: dict):
    version = header.get('version', 1)
    file_format = header.get('format', 1)
    info = header.get('info', {})
    header = lzma.compress(dill.dumps(info),
                           format=lzma.FORMAT_RAW,
                           filters=lzma_filters)
    assert len(header) < 2**16, 'Header too large'
    version = version.to_bytes(1, 'big', signed=False)
    file_format = file_format.to_bytes(1, 'big', signed=False)
    info_size = len(header).to_bytes(2, 'big', signed=False)
    return MAGIC + version + file_format + info_size + header


class FakeLock():

    def __enter__(self):
        pass

    def __exit__(self, *args):
        pass

    async def __aenter__(self):
        pass

    async def __aexit__(self, *args):
        pass


class BaseFile():

    def __init__(self, path: str | pathlib.Path, lock=None):
        self.path = path
        self.lock = lock or FakeLock()

    def compress(self):
        with self.lock:
            self._compress()

    def decompress(self):
        with self.lock:
            self._decompress()

    def _compress(self):
        with open(self.path, 'rb+') as f:
            try:
                header = _get_header(f)
            except:
                return
            if header.get('format', FORMAT_RAW) == FORMAT_XZ:
                return
            header['format'] = FORMAT_XZ
            buffer = f.read()
            f.seek(0)
            f.write(_make_header(header))
            f.write(
                lzma.compress(buffer,
                              format=lzma.FORMAT_RAW,
                              filters=lzma_filters))
            f.truncate()

    def _decompress(self):
        with open(self.path, 'rb+') as f:
            try:
                header = _get_header(f)
            except:
                return
            if header.get('format', FORMAT_RAW) == FORMAT_RAW:
                return
            header['format'] = FORMAT_RAW
            buffer = lzma.decompress(f.read(),
                                     format=lzma.FORMAT_RAW,
                                     filters=lzma_filters)
            f.seek(0)
            f.write(_make_header(header))
            f.write(buffer)
            f.truncate()


class BinaryFile(BaseFile):

    def read(self):
        with self.lock:
            with open(self.path, 'rb') as f:
                header = _get_header(f)
                if header.get('format', FORMAT_RAW) == FORMAT_XZ:
                    return lzma.decompress(f.read(),
                                           format=lzma.FORMAT_RAW,
                                           filters=lzma_filters)
                else:
                    return f.read()

    def write(self, data: bytes):
        with self.lock:
            with open(self.path, 'wb') as f:
                f.write(
                    _make_header({
                        'format': FORMAT_RAW,
                        'info': {
                            'type': 'bytes'
                        }
                    }))
                f.write(data)


class ObjectFile(BinaryFile):

    def load(self):
        with self.lock:
            with open(self.path, 'rb') as f:
                header = _get_header(f)
                if header.get('format', FORMAT_RAW) == FORMAT_XZ:
                    return dill.loads(
                        lzma.decompress(f.read(),
                                        format=lzma.FORMAT_RAW,
                                        filters=lzma_filters))
                else:
                    return dill.load(f)

    def dump(self, obj):
        with self.lock:
            with open(self.path, 'wb') as f:
                f.write(
                    _make_header({
                        'format': FORMAT_RAW,
                        'info': {
                            'type': 'pickle'
                        }
                    }))
                dill.dump(obj, f)


class ObjectListFile(ObjectFile):

    def clear(self):
        with self.lock:
            self._decompress()
            with open(self.path, 'wb') as f:
                f.write(
                    _make_header({
                        'format': FORMAT_RAW,
                        'info': {
                            'type': 'pickle_list'
                        }
                    }))

    def append(self, obj):
        with self.lock:
            with open(self.path, 'ab') as f:
                if f.tell() == 0:
                    f.write(
                        _make_header({
                            'format': FORMAT_RAW,
                            'info': {
                                'type': 'pickle_list'
                            }
                        }))
                dill.dump(obj, f)

    def __iter__(self):
        with self.lock:
            with open(self.path, 'rb') as f:
                while True:
                    try:
                        header = _get_header(f)
                    except AssertionError:
                        break
                    if header.get('format', FORMAT_RAW) == FORMAT_XZ:
                        yield from self._compressed_iter(f)
                    else:
                        yield from self._iter(f)

    def _compressed_iter(self, f):
        with lzma.open(f, 'rb', format=lzma.FORMAT_RAW,
                       filters=lzma_filters) as f:
            yield from self._iter(f)

    def _iter(self, f):
        while True:
            try:
                yield dill.load(f)
            except EOFError:
                break

    def asarray(self):
        return np.array(list(self))


class ArrayFile(BaseFile):

    def clear(self):
        with self.lock:
            with open(self.path, 'wb') as f:
                f.write(b'')

    def extend(self, data: np.ndarray):
        with self.lock:
            self._decompress()
            with open(self.path, 'ab') as f:
                if f.tell() == 0:
                    f.write(
                        _make_header({
                            'format': FORMAT_RAW,
                            'info': {
                                'type': 'array',
                                'dtype': data.dtype
                            }
                        }))
                f.write(data.tobytes())

    def append(self, data, dtype=float):
        self.extend(np.asarray(data, dtype=dtype))

    def asarray(self):
        with self.lock:
            with open(self.path, 'rb') as f:
                header = _get_header(f)
                dtype = header['info']['dtype']
                if header.get('format', FORMAT_RAW) == FORMAT_XZ:
                    return np.frombuffer(
                        lzma.decompress(f.read(),
                                        format=lzma.FORMAT_RAW,
                                        filters=lzma_filters),
                        dtype=dtype)
                else:
                    return np.fromfile(f, dtype=dtype)


def load(path):
    with open(path, 'rb') as f:
        header = _get_header(f)
        if header['info']['type'] == 'array':
            return ArrayFile(path)
        elif header['info']['type'] == 'pickle':
            return ObjectFile(path)
        elif header['info']['type'] == 'pickle_list':
            return ObjectListFile(path)
        elif header['info']['type'] == 'bytes':
            return BinaryFile(path)
        else:
            raise ValueError(f'Unknown file type: {header["info"]["type"]}')



import os

from Cython.Build import cythonize
from setuptools import Extension, find_packages, setup


def module_name(dirpath, filename):
    l = [os.path.splitext(filename)[0]]
    head = dirpath
    while True:
        head, tail = os.path.split(head)
        l.append(tail)
        if not head:
            break
    return '.'.join(reversed(l))


def get_extensions():
    extensions = []
    for dirpath, dirnames, filenames in os.walk('qulab'):
        for filename in filenames:
            if filename.endswith('.pyx'):
                extensions.append(
                    Extension(module_name(dirpath, filename),
                              [os.path.join(dirpath, filename)]))

    return extensions


setup(packages=find_packages(exclude=['tests']),
      ext_modules=cythonize(get_extensions(), build_dir=f"build"))



"""
This module provides a simple bencoding implementation in pure Python.
Bencoding is a way to specify and organize data in a terse format. It supports
the following types: byte strings, integers, lists, and dictionaries.

Classes:
    Bencached: A class that holds bencoded data.

Functions:
    encode_bencached(x, fp): Encodes a Bencached object.
    encode_int(x, fp): Encodes an integer.
    encode_bool(x, fp): Encodes a boolean.
    encode_string(x, fp): Encodes a string.
    encode_list(x, fp): Encodes a list.
    encode_dict(x, fp): Encodes a dictionary.
    decode_int(x, ptr): Decodes a bencoded integer.
    decode_string(x, ptr): Decodes a bencoded string.
    decode_list(x, ptr): Decodes a bencoded list.
    decode_dict(x, ptr): Decodes a bencoded dictionary.
    encode(obj, fp): Encodes an object.
    encodes(obj): Encodes an object and returns the bencoded bytes.
    decode(fp): Decodes a file-like object.
    decodes(s): Decodes a bencoded bytes object.
    dump = encode: Alias for encode function.
    load = decodes: Alias for decode function.
    dumps = encodes: Alias for encodes function.
    loads = decodes: Alias for decodes function.

Examples:
    >>> encode(42) == b'i42e'
    True
    >>> decode(b'i42e')
    42
"""
import io


class Bencached():

    __slots__ = ['bencoded']

    def __init__(self, s):
        self.bencoded = s


def encode_bencached(x, fp):
    fp.write(x.bencoded)


def encode_int(x, fp):
    fp.write(f"i{x}e".encode())


def encode_bool(x, fp):
    if x:
        encode_int(1, fp)
    else:
        encode_int(0, fp)


def encode_string(x, fp):
    if isinstance(x, str):
        x = x.encode()
    fp.write(f"{len(x):d}:".encode())
    fp.write(x)


def encode_list(x, fp):
    fp.write(b'l')
    for i in x:
        encode_func[type(i)](i, fp)
    fp.write(b'e')


def encode_dict(x, fp):
    fp.write(b'd')
    for k, v in sorted(x.items()):
        if isinstance(k, str):
            k = k.encode()
        fp.write(f"{len(k):d}:".encode())
        fp.write(k)
        encode_func[type(v)](v, fp)
    fp.write(b'e')


encode_func = {
    Bencached: encode_bencached,
    int: encode_int,
    str: encode_string,
    bytes: encode_string,
    list: encode_list,
    tuple: encode_list,
    dict: encode_dict,
    bool: encode_bool
}


def decode_int(x, ptr):
    ptr += 1
    end = x.index(b'e', ptr)
    if x[ptr] == b'-'[0]:
        if x[ptr + 1] == b'0'[0]:
            raise ValueError
    elif x[ptr] == b'0'[0] and end != ptr + 1:
        raise ValueError
    n = int(x[ptr:end])
    return n, end + 1


def decode_string(x, ptr):
    colon = x.index(b':', ptr)
    n = int(x[ptr:colon])
    if x[ptr] == b'0'[0] and colon != ptr + 1:
        raise ValueError
    colon += 1
    return x[colon:colon + n], colon + n


def decode_list(x, ptr):
    r, ptr = [], ptr + 1
    while x[ptr] != b'e'[0]:
        v, ptr = decode_func[x[ptr]](x, ptr)
        r.append(v)
    return r, ptr + 1


def decode_dict(x, ptr):
    r, ptr = {}, ptr + 1
    while x[ptr] != b'e'[0]:
        k, ptr = decode_string(x, ptr)
        if isinstance(k, bytes):
            k = k.decode()
        r[k], ptr = decode_func[x[ptr]](x, ptr)
    return r, ptr + 1


decode_func = {b'l'[0]: decode_list, b'd'[0]: decode_dict, b'i'[0]: decode_int}

for i in range(10):
    decode_func[f"{i}".encode()[0]] = decode_string


def encode(obj, fp):
    try:
        encode_func[type(obj)](obj, fp)
    except KeyError:
        raise ValueError("Allowed types: int, bytes, list, dict; not %s",
                         type(obj))


def encodes(obj):
    """
    bencodes given object. Given object should be a int,
    bytes, list or dict. If a str is given, it'll be
    encoded as ASCII.
    >>> [encode(i) for i in (-2, 42, b"answer", b"")] \
            == [b'i-2e', b'i42e', b'6:answer', b'0:']
    True
    >>> encode([b'a', 42, [13, 14]]) == b'l1:ai42eli13ei14eee'
    True
    >>> encode({'bar': b'spam', 'foo': 42, 'mess': [1, b'c']}) \
            == b'd3:bar4:spam3:fooi42e4:messli1e1:cee'
    True
    """
    buff = io.BytesIO()
    encode(obj, buff)
    return buff.getvalue()


def decodes(s):
    """
    Decodes given bencoded bytes object.
    >>> decode(b'i-42e')
    -42
    >>> decode(b'4:utku') == b'utku'
    True
    >>> decode(b'li1eli2eli3eeee')
    [1, [2, [3]]]
    >>> decode(b'd3:bar4:spam3:fooi42ee') == {'bar': b'spam', 'foo': 42}
    True
    """
    if isinstance(s, str):
        s = s.encode()
    try:
        r, l = decode_func[s[0]](s, 0)
    except (IndexError, KeyError, ValueError):
        raise ValueError("not a valid bencoded string")
    if l != len(s):
        raise ValueError("invalid bencoded value (data after valid prefix)")
    return r


def decode(fp):
    """
    Decodes given file-like object.
    """
    return decodes(fp.read())


dump = encode
load = decodes
dumps = encodes
loads = decodes

__all__ = ['decodes', 'encodes', 'dump', 'load', 'dumps', 'loads']



import bisect
from concurrent.futures import Future
from datetime import datetime
from itertools import chain
from multiprocessing import Lock
from queue import Queue
from typing import Any, Sequence

from ..scan.base import StepStatus, Tracker, _get_all_dependence

_NODEFAULT = object()


class BaseDataset(Tracker):
    """
    A tracker that stores the results of the steps.

    Parameters
    ----------
    data : dict
        The data of the results.
    shape : tuple
        The shape of the results.
    ctime : datetime.datetime
        The creation time of the tracker.
    mtime : datetime.datetime
        The modification time of the tracker.
    """

    def __init__(
            self,
            data: dict = None,
            shape: tuple = (),
            save_kwds: bool | Sequence[str] = True,
            frozen_keys: tuple = (),
            ignores: tuple = (),
    ):
        self.ctime = datetime.utcnow()
        self.mtime = datetime.utcnow()
        self.data = data if data is not None else {}
        self.cache = {}
        self.pos = {}
        self.timestamps = {}
        self.iteration = {}
        self._init_keys = list(self.data.keys())
        self._frozen_keys = frozen_keys
        self._ignores = ignores
        self._key_levels = ()
        self.depends = {}
        self.dims = {}
        self.vars_dims = {}
        self.shape = shape
        self.count = 0
        self.save_kwds = save_kwds
        self.queue = Queue()
        self._queue_buffer = None
        self._lock = Lock()

    def init(self, loops: dict, functions: dict, constants: dict, graph: dict,
             order: list):
        """
        Initialize the tracker.

        Parameters
        ----------
        loops : dict
            The map of iterables.
        functions : dict
            The map of functions.
        constants : dict
            The map of constants.
        graph : dict
            The dependence graph.
        order : list
            The order of the dependence graph.
        """
        from numpy import ndarray

        self.depends = graph

        for level, (keys, iters) in enumerate(loops.items()):
            self._key_levels = self._key_levels + ((keys, level), )
            if isinstance(keys, str):
                keys = (keys, )
                iters = (iters, )
            if (len(keys) > 1 and len(iters) == 1
                    and isinstance(iters[0], ndarray) and iters[0].ndim == 2
                    and iters[0].shape[1] == len(keys)):
                iters = iters[0]
                for i, key in enumerate(keys):
                    self.data[key] = iters[:, i]
                    self._frozen_keys = self._frozen_keys + (key, )
                    self._init_keys.append(key)
                continue
            if not isinstance(iters, tuple) or len(keys) != len(iters):
                continue
            for key, iter in zip(keys, iters):
                if self.depends.get(key, set()):
                    dims = set()
                    for dep in self.depends[key]:
                        if dep in self.vars_dims:
                            dims.update(set(self.vars_dims[dep]))
                    dims.add(level)
                    self.vars_dims[key] = tuple(sorted(dims))
                else:
                    self.vars_dims[key] = (level, )
                if level not in self.dims:
                    self.dims[level] = ()
                self.dims[level] = self.dims[level] + (key, )
                if key not in self.data and isinstance(iter,
                                                       (list, range, ndarray)):
                    if key.startswith('__'):
                        continue
                    self.data[key] = iter
                    self._frozen_keys = self._frozen_keys + (key, )
                    self._init_keys.append(key)

        for key, value in constants.items():
            if key.startswith('__'):
                continue
            self.data[key] = value
            self._init_keys.append(key)
            self.vars_dims[key] = ()

        for ready in order:
            for key in ready:
                if key in functions:
                    deps = _get_all_dependence(key, graph)
                    dims = set()
                    for k in deps:
                        dims.update(set(self.vars_dims.get(k, ())))
                    self.vars_dims[key] = tuple(sorted(dims))

        for k, v in self.vars_dims.items():
            if len(v) == 1:
                if v[0] in self.dims and k not in self.dims[v[0]]:
                    self.dims[v[0]] = self.dims[v[0]] + (k, )
                elif v[0] not in self.dims:
                    self.dims[v[0]] = (k, )

    def feed(self,
             step: StepStatus,
             dataframe: dict | Future,
             store=False,
             **options):
        """
        Feed the results of the step to the dataset.

        Parameters
        ----------
        step : StepStatus
            The step.
        dataframe : dict
            The results of the step.
        """
        import numpy as np

        if not store:
            return
        self.mtime = datetime.utcnow()
        if not self.shape:
            self.shape = tuple([i + 1 for i in step.pos])
        else:
            self.shape = tuple(
                [max(i + 1, j) for i, j in zip(step.pos, self.shape)])
        if self.save_kwds:
            if isinstance(self.save_kwds, bool):
                kwds = step.kwds
            else:
                kwds = {
                    key: step.kwds.get(key, np.nan)
                    for key in self.save_kwds
                }
        else:
            kwds = {}

        if isinstance(dataframe, dict):
            dataframe = self._prune(dataframe)
        self.queue.put_nowait(
            (step.iteration, step.pos, dataframe, kwds, self.mtime))
        self.flush()

    def _prune(self, dataframe: dict[str, Any]) -> dict[str, Any]:
        return {
            k: v
            for k, v in dataframe.items() if k not in self._ignores
            and k not in self._frozen_keys and not k.startswith('__')
        }

    def _append(self, iteration, pos, dataframe, kwds, now):
        for k, v in chain(kwds.items(), dataframe.items()):
            if k in self._frozen_keys or k in self._ignores:
                continue
            if k.startswith('__'):
                continue
            if self.vars_dims.get(k, ()) == () and k not in dataframe:
                continue
            self.count += 1
            if k not in self.data:
                self.data[k] = [v]
                if k in self.vars_dims:
                    self.pos[k] = tuple([pos[i]] for i in self.vars_dims[k])
                else:
                    self.pos[k] = tuple([i] for i in pos)
                self.timestamps[k] = [now.timestamp()]
                self.iteration[k] = [iteration]
            else:
                if k in self.vars_dims:
                    pos_k = tuple(pos[i] for i in self.vars_dims[k])
                    if k not in dataframe and pos_k in zip(*self.pos[k]):
                        continue
                    for i, l in zip(pos_k, self.pos[k]):
                        l.append(i)
                else:
                    for i, l in zip(pos, self.pos[k]):
                        l.append(i)
                self.timestamps[k].append(now.timestamp())
                self.iteration[k].append(iteration)
                self.data[k].append(v)

    def flush(self, block=False):
        with self._lock:
            self._flush(block=block)

    def _dataframe_done(self, dataframe: Future | dict) -> bool:
        if isinstance(dataframe, Future):
            return dataframe.done()
        else:
            return all(x.done() for x in dataframe.values()
                       if isinstance(x, Future))

    def _dataframe_result(self, dataframe: Future | dict) -> dict:
        if isinstance(dataframe, Future):
            return dataframe.result()
        else:
            return {
                k: v.result() if isinstance(v, Future) else v
                for k, v in dataframe.items()
            }

    def _flush(self, block=False):
        if self._queue_buffer is not None:
            iteration, pos, dataframe, kwds, now = self._queue_buffer
            if self._dataframe_done(dataframe) or block:
                self._append(iteration, pos,
                             self._prune(self._dataframe_result(dataframe)),
                             kwds, now)
                self._queue_buffer = None
            else:
                return
        while not self.queue.empty():
            iteration, pos, dataframe, kwds, now = self.queue.get()
            if not self._dataframe_done(dataframe) and not block:
                self._queue_buffer = (iteration, pos, dataframe, kwds, now)
                return
            else:
                self._append(iteration, pos,
                             self._prune(self._dataframe_result(dataframe)),
                             kwds, now)

    def _get_array(self, key, shape, count):
        import numpy as np

        if key in self.vars_dims:
            shape = tuple([shape[i] for i in self.vars_dims[key]])

        data, data_shape, data_count = self.cache.get(key, (None, (), 0))
        if (data_shape, data_count) == (shape, count):
            return data
        try:
            tmp = np.asarray(self.data[key])
            if data_shape != shape:
                data = np.full(shape + tmp.shape[1:], np.nan, dtype=tmp.dtype)
        except:
            tmp = self.data[key]
            if data_shape != shape:
                data = np.full(shape, np.nan, dtype=object)
        try:
            data[self.pos[key]] = tmp
        except:
            raise
        self.cache[key] = (data, shape, count)
        return data

    def _get_part(self, key, skip):
        i = bisect.bisect_left(self.iteration[key], skip)
        pos = tuple(p[i:] for p in self.pos[key])
        iteration = self.iteration[key][i:]
        data = self.data[key][i:]
        return data, iteration, pos

    def keys(self):
        """
        Get the keys of the dataset.
        """
        self.flush()
        return self.data.keys()

    def values(self):
        """
        Get the values of the dataset.
        """
        self.flush()
        return [self[k] for k in self.data]

    def items(self):
        """
        Get the items of the dataset.
        """
        self.flush()
        return list(zip(self.keys(), self.values()))

    def get(self, key, default=_NODEFAULT, skip=None, block=False):
        """
        Get the value of the dataset.
        """
        self.flush(block)
        if key in self._init_keys:
            return self.data[key]
        elif key in self.data:
            if skip is None:
                return self._get_array(key, self.shape, self.count)
            else:
                return self._get_part(key, skip)
        elif default is _NODEFAULT:
            raise KeyError(key)
        else:
            return default

    def __getitem__(self, key):
        return self.get(key)

    def __getstate__(self):
        self.flush()
        data = dict(self.items())
        return {
            'data': data,
            'pos': self.pos,
            'timestamps': self.timestamps,
            'iteration': self.iteration,
            'depends': self.depends,
            'shape': self.shape,
            'dims': self.dims,
            'vars_dims': self.vars_dims,
            'ctime': self.ctime,
            'mtime': self.mtime,
            '_init_keys': self._init_keys,
            '_frozen_keys': self._frozen_keys,
            '_ignores': self._ignores,
            '_key_levels': self._key_levels,
            'save_kwds': self.save_kwds,
        }



from __future__ import annotations

import operator

import numpy as np
from pyparsing import (CaselessKeyword, Combine, Forward, Group, Keyword,
                       Literal, Optional, ParserElement, Suppress, Word,
                       alphanums, alphas, delimitedList, nums, oneOf, opAssoc,
                       pyparsing_common, restOfLine, srange, stringEnd,
                       stringStart)
from scipy import special

LPAREN, RPAREN, LBRACK, RBRACK, LBRACE, RBRACE, DOT, TILDE, BANG, PLUS, MINUS = map(
    Suppress, "()[]{}.~!+-")

INT = Combine(srange("[1-9]") +
              Optional(Word(nums))).set_parse_action(lambda t: int(t[0]))
OCT = Combine("0" + Word("01234567")).set_parse_action(lambda t: int(t[0], 8))
HEX = Combine("0x" + Word("0123456789abcdefABCDEF")).set_parse_action(
    lambda t: int(t[0], 16))
FLOAT = Combine(Word(nums) + DOT + Word(nums)) | \
        Combine(DOT + Word(nums)) | \
        Combine(Word(nums) + DOT) | \
        Combine(Word(nums) + DOT + Word(nums) + CaselessKeyword("e") + Word("+-") + Word(nums)) | \
        Combine(Word(nums) + DOT + CaselessKeyword("e") + Word("+-") + Word(nums)) | \
        Combine(DOT + Word(nums) + CaselessKeyword("e") + Word("+-") + Word(nums)) | \
        Combine(Word(nums) + CaselessKeyword("e") + Word("+-") + Word(nums))
FLOAT.set_parse_action(lambda t: float(t[0]))
SYMBOL = Word(alphas, alphanums + "_")
SYMBOL.set_parse_action(lambda t: Symbol(t[0]))

expr = Forward()
unary = Forward()
binary = Forward()
atom = Forward()

atom << (INT | OCT | HEX | FLOAT | SYMBOL | (LPAREN + expr + RPAREN) |
         (LBRACK + expr + RBRACK) | (LBRACE + expr + RBRACE) | (MINUS + atom) |
         (PLUS + atom) | (TILDE + atom) | (BANG + atom) | (nums + DOT + nums))

unary << (atom | (MINUS + unary) | (PLUS + unary) | (TILDE + unary) |
          (BANG + unary))

ConstType = (int, float, complex)
_empty = object()


class Ref():
    __slots__ = ['name']

    def __init__(self, name):
        self.name = name

    def __repr__(self) -> str:
        return f"Ref({self.name!r})"


class Env():

    def __init__(self):
        self.consts = {}
        self.variables = {}
        self.refs = {}
        self.functions = {
            'sin': np.sin,
            'cos': np.cos,
            'tan': np.tan,
            'pi': np.pi,
            'e': np.e,
            'log': np.log,
            'log2': np.log2,
            'log10': np.log10,
            'exp': np.exp,
            'sqrt': np.sqrt,
            'abs': np.abs,
            'sinh': np.sinh,
            'cosh': np.cosh,
            'tanh': np.tanh,
            'arcsin': np.arcsin,
            'arccos': np.arccos,
            'arctan': np.arctan,
            'arctan2': np.arctan2,
            'arcsinh': np.arcsinh,
            'arccosh': np.arccosh,
            'arctanh': np.arctanh,
            'sinc': np.sinc,
            'sign': np.sign,
            'heaviside': np.heaviside,
            'erf': special.erf,
            'erfc': special.erfc,
        }

    def __contains__(self, key):
        return key in self.consts or key in self.variables or key in self.functions or key in self.refs

    def __getitem__(self, key):
        if key in self.consts:
            return self.consts[key]
        if key in self.variables:
            return self.variables[key]
        if key in self.functions:
            return self.functions[key]
        if key in self.refs:
            return self[self.refs[key]]
        raise KeyError(f"Key {key} not found")

    def __setitem__(self, key, value):
        if key in self.consts:
            raise KeyError(f"Key {key:r} is const")
        elif isinstance(value, Ref):
            self.create_ref(key, value.name)
        elif key in self.refs:
            self[self.refs[key]] = value
        else:
            self.variables[key] = value

    def __delitem__(self, key):
        if key in self.consts:
            raise KeyError(f"Key {key:r} is const")
        elif key in self.refs:
            del self[self.refs[key]]
        else:
            del self.variables[key]

    def ref(self, key):
        if key in self:
            return Ref(key)
        else:
            raise KeyError(f"Key {key!r} not found")

    def create_ref(self, key, name):
        if name in self.refs:
            if key in self.refs[name]:
                raise ValueError(f"Key {key!r} already exists in ref {name!r}")
            else:
                self.refs[key] = [name, *self.refs[name]]
        else:
            self.refs[key] = [name]

    def is_const(self, key):
        return key in self.consts


_default_env = Env()


class Expression():

    def __init__(self):
        self.cache = _empty

    def d(self, x: str | Symbol):
        if isinstance(x, Symbol):
            x = x.name
        if x in self.symbols():
            return self.derivative(x)
        else:
            return 0

    def derivative(self, x):
        raise NotImplementedError

    def __add__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        if isinstance(other, ConstType) and other == 0:
            return self
        return BinaryExpression(self, other, operator.add)

    def __radd__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        if isinstance(other, ConstType) and other == 0:
            return self
        return BinaryExpression(other, self, operator.add)

    def __sub__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        if isinstance(other, ConstType) and other == 0:
            return self
        return BinaryExpression(self, other, operator.sub)

    def __rsub__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        if isinstance(other, ConstType) and other == 0:
            return -self
        return BinaryExpression(other, self, operator.sub)

    def __mul__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        if isinstance(other, ConstType) and other == 0:
            return 0
        if isinstance(other, ConstType) and other == 1:
            return self
        if isinstance(other, ConstType) and other == -1:
            return -self
        return BinaryExpression(self, other, operator.mul)

    def __rmul__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        if isinstance(other, ConstType) and other == 0:
            return 0
        if isinstance(other, ConstType) and other == 1:
            return self
        if isinstance(other, ConstType) and other == -1:
            return -self
        return BinaryExpression(other, self, operator.mul)

    def __matmul__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        return BinaryExpression(self, other, operator.matmul)

    def __rmatmul__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        return BinaryExpression(other, self, operator.matmul)

    def __truediv__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        if isinstance(other, ConstType) and other == 1:
            return self
        if isinstance(other, ConstType) and other == -1:
            return -self
        return BinaryExpression(self, other, operator.truediv)

    def __rtruediv__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        if isinstance(other, ConstType) and other == 0:
            return 0
        return BinaryExpression(other, self, operator.truediv)

    def __floordiv__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        if isinstance(other, ConstType) and other == 1:
            return self
        if isinstance(other, ConstType) and other == -1:
            return -self
        return BinaryExpression(self, other, operator.floordiv)

    def __rfloordiv__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        if isinstance(other, ConstType) and other == 0:
            return 0
        return BinaryExpression(other, self, operator.floordiv)

    def __mod__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        if isinstance(other, ConstType) and other == 1:
            return 0
        return BinaryExpression(self, other, operator.mod)

    def __rmod__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        return BinaryExpression(other, self, operator.mod)

    def __pow__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        if isinstance(other, ConstType) and other == 0:
            return 1
        if isinstance(other, ConstType) and other == 1:
            return self
        return BinaryExpression(self, other, operator.pow)

    def __rpow__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        if isinstance(other, ConstType) and other == 0:
            return 0
        return BinaryExpression(other, self, operator.pow)

    def __neg__(self):
        return UnaryExpression(self, operator.neg)

    def __pos__(self):
        return UnaryExpression(self, operator.pos)

    def __abs__(self):
        return UnaryExpression(self, operator.abs)

    def __not__(self):
        return UnaryExpression(self, operator.not_)

    def __inv__(self):
        return UnaryExpression(self, operator.inv)

    def __invert__(self):
        return UnaryExpression(self, operator.invert)

    def __index__(self):
        return UnaryExpression(self, operator.index)

    def __eq__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        return BinaryExpression(self, other, operator.eq)

    def __ne__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        return BinaryExpression(self, other, operator.ne)

    def __lt__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        return BinaryExpression(self, other, operator.lt)

    def __le__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        return BinaryExpression(self, other, operator.le)

    def __gt__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        return BinaryExpression(self, other, operator.gt)

    def __ge__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        return BinaryExpression(self, other, operator.ge)

    def __and__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        return BinaryExpression(self, other, operator.and_)

    def __rand__(self, other):
        if isinstance(other, Expression):
            other
        return BinaryExpression(other, self, operator.and_)

    def __or__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        return BinaryExpression(self, other, operator.or_)

    def __ror__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        return BinaryExpression(other, self, operator.or_)

    def __lshift__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        return BinaryExpression(self, other, operator.lshift)

    def __rlshift__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        return BinaryExpression(other, self, operator.lshift)

    def __rshift__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        return BinaryExpression(self, other, operator.rshift)

    def __rrshift__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        return BinaryExpression(other, self, operator.rshift)

    def __xor__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        return BinaryExpression(self, other, operator.xor)

    def __rxor__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        return BinaryExpression(other, self, operator.xor)

    def __getitem__(self, other):
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        return ObjectMethod(self, '__getitem__', other)

    def __getattr__(self, other):
        if isinstance(other, str):
            if other.startswith('_') or other in self.__dict__:
                return super().__getattr__(other)
        if isinstance(other, Expression):
            other = other.eval(_default_env)
        return ObjectMethod(self, '__getattr__', other)

    def __call__(self, *args):
        args = [
            o.eval(_default_env) if isinstance(o, Expression) else o
            for o in args
        ]
        return ObjectMethod(self, '__call__', *args)

    def __round__(self, n=None):
        return self

    def __bool__(self):
        return True

    def eval(self, env):
        raise NotImplementedError

    def symbols(self) -> list[str]:
        raise NotImplementedError

    def changed(self, env) -> bool:
        return True

    def is_const(self, env) -> bool:
        return False

    def value(self, env=_default_env):
        if isinstance(env, dict):
            e = Env()
            e.variables = env
            env = e
        if self.changed(env):
            self.cache = self.eval(env)
        return self.cache


class UnaryExpression(Expression):

    def __init__(self, a, op):
        super().__init__()
        self.a = a
        self.op = op

    def __getstate__(self) -> dict:
        return {'a': self.a, 'op': self.op}

    def __setstate__(self, state: dict):
        self.a = state['a']
        self.op = state['op']
        self.cache = _empty

    def symbols(self) -> list[str]:
        if isinstance(self.a, Expression):
            return self.a.symbols()
        else:
            return []

    def changed(self, env) -> bool:
        if isinstance(self.a, ConstType):
            return False
        return self.cache is _empty or isinstance(
            self.a, Expression) and self.a.changed(env)

    def is_const(self, env) -> bool:
        return isinstance(self.a,
                          Expression) and self.a.is_const(env) or isinstance(
                              self.a, ConstType)

    def eval(self, env):
        a = self.a.value(env) if isinstance(self.a, Expression) else self.a
        return self.op(a)

    def derivative(self, x):
        if isinstance(self.a, Expression):
            return self.op(self.a.d(x))
        else:
            return 0

    def __repr__(self) -> str:
        return f"{self.op.__name__}({self.a!r})"


class BinaryExpression(Expression):

    def __init__(self, a, b, op):
        super().__init__()
        self.a = a
        self.b = b
        self.op = op

    def __getstate__(self) -> dict:
        return {'a': self.a, 'b': self.b, 'op': self.op}

    def __setstate__(self, state: dict):
        self.a = state['a']
        self.b = state['b']
        self.op = state['op']
        self.cache = _empty

    def symbols(self) -> list[str]:
        symbs = set()
        if isinstance(self.a, Expression):
            symbs.update(self.a.symbols())
        if isinstance(self.b, Expression):
            symbs.update(self.b.symbols())
        return list(symbs)

    def eval(self, env):
        a = self.a.value(env) if isinstance(self.a, Expression) else self.a
        b = self.b.value(env) if isinstance(self.b, Expression) else self.b
        return self.op(a, b)

    def derivative(self, x):
        if isinstance(self.a, Expression):
            da = self.a.d(x)
        else:
            da = 0
        if isinstance(self.b, Expression):
            db = self.b.d(x)
        else:
            db = 0

        if self.op is operator.add:
            return da + db
        elif self.op is operator.sub:
            return da - db
        elif self.op is operator.mul:
            return self.a * db + da * self.b
        elif self.op is operator.truediv:
            return (da * self.b - self.a * db) / self.b**2
        elif self.op is operator.pow:
            if isinstance(self.a, Expression) and isinstance(
                    self.b, Expression):
                return self.a**self.b * (self.b * da / self.a +
                                         ObjectMethod(np, 'log', self.a) * db)
            elif isinstance(self.a, Expression):
                return self.b * self.a**(self.b - 1) * da
            elif isinstance(self.b, Expression):
                return np.log(self.a) * db * self.a**self.b
            else:
                return 0
        else:
            return 0

    def __repr__(self) -> str:
        return f"({self.a!r} {self.op.__name__} {self.b!r})"


class ObjectMethod(Expression):

    def __init__(self, obj, method: str, *args):
        super().__init__()
        self.obj = obj
        self.method = method
        self.args = args

    def __getstate__(self) -> dict:
        return {'obj': self.obj, 'method': self.method, 'args': self.args}

    def __setstate__(self, state: dict):
        self.obj = state['obj']
        self.method = state['method']
        self.args = state['args']
        self.cache = _empty

    def symbols(self) -> list[str]:
        symbs = set()
        if isinstance(self.obj, Expression):
            symbs.update(self.obj.symbols())
        for a in self.args:
            if isinstance(a, Expression):
                symbs.update(a.symbols())
        return list(symbs)

    def eval(self, env):
        obj = self.obj.value(env) if isinstance(self.obj,
                                                Expression) else self.obj
        args = [
            a.value(env) if isinstance(a, Expression) else a for a in self.args
        ]
        if isinstance(obj, Expression) or any(
                isinstance(x, Expression) for x in args):
            return ObjectMethod(obj, self.method, *args)
        else:
            return getattr(obj, self.method)(*args)

    def __repr__(self):
        if self.method == '__call__':
            return f"{self.obj!r}({', '.join(map(repr, self.args))})"
        else:
            return f"{self.obj!r}.{self.method}({', '.join(map(repr, self.args))})"


class Symbol(Expression):

    def __init__(self, name):
        super().__init__()
        self.name = name

    def __getstate__(self) -> dict:
        return {'name': self.name}

    def __setstate__(self, state: dict):
        self.name = state['name']
        self.cache = _empty

    def symbols(self) -> list[str]:
        return [self.name]

    def eval(self, env):
        if self.name in env:
            return env[self.name]
        else:
            return self

    def derivative(self, x):
        if x == self.name:
            return 1
        else:
            return 0

    def __repr__(self) -> str:
        return self.name


sin = Symbol('sin')
cos = Symbol('cos')
tan = Symbol('tan')
pi = Symbol('pi')
e = Symbol('e')
log = Symbol('log')
log2 = Symbol('log2')
log10 = Symbol('log10')
exp = Symbol('exp')
sqrt = Symbol('sqrt')
abs = Symbol('abs')
sinh = Symbol('sinh')
cosh = Symbol('cosh')
tanh = Symbol('tanh')
arcsin = Symbol('arcsin')
arccos = Symbol('arccos')
arctan = Symbol('arctan')
arctan2 = Symbol('arctan2')
arcsinh = Symbol('arcsinh')
arccosh = Symbol('arccosh')
arctanh = Symbol('arctanh')
sinc = Symbol('sinc')
sign = Symbol('sign')
heaviside = Symbol('heaviside')
erf = Symbol('erf')
erfc = Symbol('erfc')
